@@ -354,34 +354,6 @@ static void HdmiSendEnable(NVDpyEvoPtr pDpyEvo, NvBool hdmiEnable)
     }
 }
 
-/*!
- * Disable sending the vendor specific infoframe on this display.
- */
-static void DisableVendorSpecificInfoFrame(
-    const NVDispEvoRec *pDispEvo,
-    const NvU32 head)
-{
-    const NVDispHeadStateEvoRec *pHeadState = &pDispEvo->headState[head];
-    NV0073_CTRL_SPECIFIC_SET_OD_PACKET_CTRL_PARAMS params = { 0 };
-    NVDevEvoPtr pDevEvo = pDispEvo->pDevEvo;
-    NvU32 ret;
-
-    params.subDeviceInstance = pDispEvo->displayOwner;
-    params.displayId = pHeadState->activeRmId;
-    params.type = pktType_VendorSpecInfoFrame;
-    params.transmitControl = DRF_DEF(0073_CTRL_SPECIFIC, _SET_OD_PACKET_CTRL_TRANSMIT_CONTROL, _ENABLE, _NO);
-
-    ret = nvRmApiControl(nvEvoGlobal.clientHandle,
-                         pDevEvo->displayCommonHandle,
-                         NV0073_CTRL_CMD_SPECIFIC_SET_OD_PACKET_CTRL,
-                         &params,
-                         sizeof(params));
-
-    if (ret != NVOS_STATUS_SUCCESS) {
-        nvAssert(!"NV0073_CTRL_CMD_SPECIFIC_SET_OD_PACKET_CTRL failed");
-    }
-}
-
 /*!
  * Sends General Control Packet to the HDMI sink.
  */
@@ -417,126 +389,9 @@ static void SendHdmiGcp(const NVDispEvoRec *pDispEvo,
     }
 }
 
-/*
- * SendInfoFrame() - Send infoframe to the hardware through the hdmipkt
- * library.
- */
-
-static void SendInfoFrame(const NVDispEvoRec *pDispEvo,
-                          const NvU32 head,
-                          NVHDMIPKT_TC transmitControl,
-                          NVT_INFOFRAME_HEADER *pInfoFrameHeader,
-                          NvU32 infoframeSize)
-{
-    const NVDispHeadStateEvoRec *pHeadState = &pDispEvo->headState[head];
-    NVDevEvoPtr pDevEvo = pDispEvo->pDevEvo;
-    NVHDMIPKT_TYPE hdmiLibType;
-    NVHDMIPKT_RESULT ret;
-    NvU8 *infoframe = NULL;
-    NvU8 hdmiPacketType, checksum;
-    NvU32 i;
-    NvU8 *pPayload;
-    size_t headerSize;
-    NvBool needChecksum =
-        (transmitControl & DRF_DEF(_HDMI_PKT, _TRANSMIT_CTRL, _CHKSUM_HW, _EN));
-
-    /*
-     * The 'type' the timing library writes into the NVT_INFOFRAME_HEADER
-     * structure is not the same as the protocol values that hardware expects
-     * to see in the real packet header; those are defined in the
-     * HDMI_PACKET_TYPE enums (hdmi_pktType_*) from hdmi_spec.h; use those
-     * to fill in the first byte of the packet.  It's *also* not the type that
-     * the HDMI library expects to see in its NvHdmiPkt_PacketWrite call; those
-     * are NVHDMIPKT_TYPE_*.  Determine both below.
-     */
-    switch (pInfoFrameHeader->type) {
-        default:
-            nvAssert(0);
-            return;
-        case NVT_INFOFRAME_TYPE_EXTENDED_METADATA_PACKET:
-            hdmiLibType = NVHDMIPKT_TYPE_GENERIC;
-            hdmiPacketType = hdmi_pktType_ExtendedMetadata;
-            break;
-        case NVT_INFOFRAME_TYPE_VIDEO:
-            hdmiLibType = NVHDMIPKT_TYPE_AVI_INFOFRAME;
-            hdmiPacketType = hdmi_pktType_AviInfoFrame;
-            break;
-        case NVT_INFOFRAME_TYPE_VENDOR_SPECIFIC:
-            hdmiLibType = NVHDMIPKT_TYPE_VENDOR_SPECIFIC_INFOFRAME;
-            hdmiPacketType = hdmi_pktType_VendorSpecInfoFrame;
-            break;
-        case NVT_INFOFRAME_TYPE_DYNAMIC_RANGE_MASTERING:
-            hdmiLibType = NVHDMIPKT_TYPE_VENDOR_SPECIFIC_INFOFRAME;
-            hdmiPacketType = hdmi_pktType_DynamicRangeMasteringInfoFrame;
-            break;
-    }
-
-    /*
-     * These structures are weird. The NVT_VIDEO_INFOFRAME,
-     * NVT_VENDOR_SPECIFIC_INFOFRAME, NVT_EXTENDED_METADATA_PACKET_INFOFRAME,
-     * etc structures are *kind of* what we want to send to the hdmipkt library,
-     * except the type in the header is different, and a single checksum byte
-     * may need to be inserted *between* the header and the payload (requiring
-     * us to allocate a buffer one byte larger).
-     */
-    infoframe = nvAlloc(infoframeSize + (needChecksum ? sizeof(checksum) : 0));
-    if (infoframe == NULL) {
-        return;
-    }
-
-    /*
-     * The fields and size of NVT_EXTENDED_METADATA_PACKET_INFOFRAME_HEADER
-     * match with those of NVT_INFOFRAME_HEADER at the time of writing, but
-     * nvtiming.h declares them separately. To be safe, special case
-     * NVT_INFOFRAME_TYPE_EXTENDED_METADATA_PACKET.
-     */
-    if (pInfoFrameHeader->type == NVT_INFOFRAME_TYPE_EXTENDED_METADATA_PACKET) {
-        NVT_EXTENDED_METADATA_PACKET_INFOFRAME_HEADER *pExtMetadataHeader =
-            (NVT_EXTENDED_METADATA_PACKET_INFOFRAME_HEADER *) pInfoFrameHeader;
-
-        pPayload = (NvU8 *)(pExtMetadataHeader + 1);
-        headerSize = sizeof(NVT_EXTENDED_METADATA_PACKET_INFOFRAME_HEADER);
-    } else {
-        pPayload = (NvU8 *)(pInfoFrameHeader + 1);
-        headerSize = sizeof(NVT_INFOFRAME_HEADER);
-    }
-
-    infoframe[0] = hdmiPacketType;
-    nvkms_memcpy(&infoframe[1], &((NvU8*) pInfoFrameHeader)[1], headerSize - 1);
-
-    if (needChecksum) {
-        /* PB0: checksum */
-        checksum = 0;
-        infoframe[headerSize] = 0;
-        for (i = 0; i < infoframeSize + sizeof(checksum); i++) {
-            checksum += infoframe[i];
-        }
-        infoframe[headerSize] = ~checksum + 1;
-    }
-
-    /* copy the payload, starting after the 3-byte header and checksum */
-    nvkms_memcpy(&infoframe[headerSize + (needChecksum ? sizeof(checksum) : 0)],
-                 pPayload, infoframeSize - headerSize /* payload size */);
-
-    ret = NvHdmiPkt_PacketWrite(pDevEvo->hdmiLib.handle,
-                                pDispEvo->displayOwner,
-                                pHeadState->activeRmId,
-                                head,
-                                hdmiLibType,
-                                transmitControl,
-                                infoframeSize,
-                                infoframe);
-
-    if (ret != NVHDMIPKT_SUCCESS) {
-        nvAssert(ret == NVHDMIPKT_SUCCESS);
-    }
-
-    nvFree(infoframe);
-}
-
 /*
  * SendVideoInfoFrame() - Construct video infoframe using provided EDID and call
- * SendInfoFrame() to send it to RM.
+ * ->SendHdmiInfoFrame() to send it to RM.
  */
 static void SendVideoInfoFrame(const NVDispEvoRec *pDispEvo,
                                const NvU32 head,
@@ -544,6 +399,7 @@ static void SendVideoInfoFrame(const NVDispEvoRec *pDispEvo,
                                const NVDispHeadInfoFrameStateEvoRec *pInfoFrameState,
                                NVT_EDID_INFO *pEdidInfo)
 {
+    const NVDevEvoRec *pDevEvo = pDispEvo->pDevEvo;
     NvBool hdTimings = pInfoFrameState->hdTimings;
     NVT_VIDEO_INFOFRAME_CTRL videoCtrl = pInfoFrameState->ctrl;
     NVT_VIDEO_INFOFRAME VideoInfoFrame;
@@ -560,46 +416,41 @@ static void SendVideoInfoFrame(const NVDispEvoRec *pDispEvo,
         return;
     }
 
-    SendInfoFrame(pDispEvo,
-                  head,
-                  NVHDMIPKT_TRANSMIT_CONTROL_ENABLE_EVERY_FRAME,
-                  (NVT_INFOFRAME_HEADER *) &VideoInfoFrame,
-                  sizeof(VideoInfoFrame));
+    pDevEvo->hal->SendHdmiInfoFrame(
+        pDispEvo,
+        head,
+        NV_EVO_INFOFRAME_TRANSMIT_CONTROL_EVERY_FRAME,
+        (NVT_INFOFRAME_HEADER *) &VideoInfoFrame,
+        (/* header length */ sizeof(NVT_INFOFRAME_HEADER) +
+         /* payload length */ VideoInfoFrame.length),
+        TRUE /* needChecksum */);
 }
 
 /*
- * SendHDMI3DVendorSpecificInfoFrame() - Construct vendor specific infoframe
- * using provided EDID and call SendInfoFrame() to send it to RM. Currently
- * hardcoded to send the infoframe necessary for HDMI 3D.
+ * SendVendorSpecificInfoFrame() - Construct vendor specific infoframe using
+ * provided EDID and call ->SendHdmiInfoFrame() to send it to RM.
  */
 
 static void
-SendHDMI3DVendorSpecificInfoFrame(const NVDispEvoRec *pDispEvo,
-                                  const NvU32 head, NVT_EDID_INFO *pEdidInfo)
+SendVendorSpecificInfoFrame(const NVDispEvoRec *pDispEvo,
+                            const NvU32 head,
+                            const NVDispHeadInfoFrameStateEvoRec *pInfoFrameState,
+                            NVT_EDID_INFO *pEdidInfo)
 {
-    const NVDispHeadStateEvoRec *pHeadState =
-                                 &pDispEvo->headState[head];
-    NVT_VENDOR_SPECIFIC_INFOFRAME_CTRL vendorCtrl = {
-        .Enable          = 1,
-        .HDMIFormat      = NVT_HDMI_VS_BYTE4_HDMI_VID_FMT_3D,
-        .HDMI_VIC        = NVT_HDMI_VS_BYTE5_HDMI_VIC_NA,
-        .ThreeDStruc     = NVT_HDMI_VS_BYTE5_HDMI_3DS_FRAMEPACK,
-        .ThreeDDetail    = NVT_HDMI_VS_BYTE_OPT1_HDMI_3DEX_NA,
-        .MetadataPresent = 0,
-        .MetadataType    = NVT_HDMI_VS_BYTE_OPT2_HDMI_METADATA_TYPE_NA,
-    };
+    const NVDevEvoRec *pDevEvo = pDispEvo->pDevEvo;
+    NVT_VENDOR_SPECIFIC_INFOFRAME_CTRL vendorCtrl = pInfoFrameState->vendorSpecificCtrl;
     NVT_VENDOR_SPECIFIC_INFOFRAME vendorInfoFrame;
     NVT_STATUS status;
 
-    if (!pEdidInfo->HDMI3DSupported) {
-        // Only send the HDMI 3D infoframe if the display supports HDMI 3D
-        return;
-    }
-
-    // Send the infoframe with HDMI 3D configured if we're setting an HDMI 3D
-    // mode.
-    if (!pHeadState->timings.hdmi3D) {
-        DisableVendorSpecificInfoFrame(pDispEvo, head);
+    /*
+     * Disable the vendor specific infoframe if not requested to be
+     * enabled, or if HDMI 3D is requested but not supported by the monitor.
+     */
+    if (!vendorCtrl.Enable ||
+        ((vendorCtrl.HDMIFormat == NVT_HDMI_VS_BYTE4_HDMI_VID_FMT_3D) &&
+         !pEdidInfo->HDMI3DSupported)) {
+        pDevEvo->hal->DisableHdmiInfoFrame(pDispEvo, head,
+                                           NVT_INFOFRAME_TYPE_VENDOR_SPECIFIC);
         return;
     }
 
@@ -612,23 +463,28 @@ SendHDMI3DVendorSpecificInfoFrame(const NVDispEvoRec *pDispEvo,
         return;
     }
 
-    SendInfoFrame(pDispEvo,
-                  head,
-                  NVHDMIPKT_TRANSMIT_CONTROL_ENABLE_EVERY_FRAME,
-                  &vendorInfoFrame.Header,
-                  sizeof(vendorInfoFrame));
+    pDevEvo->hal->SendHdmiInfoFrame(
+        pDispEvo,
+        head,
+        NV_EVO_INFOFRAME_TRANSMIT_CONTROL_EVERY_FRAME,
+        &vendorInfoFrame.Header,
+        (/* header length */ sizeof(vendorInfoFrame.Header) +
+         /* payload length */ vendorInfoFrame.Header.length),
+        TRUE /* needChecksum */);
 }
 
 static void
 SendHDRInfoFrame(const NVDispEvoRec *pDispEvo, const NvU32 head,
                  NVT_EDID_INFO *pEdidInfo)
 {
+    const NVDevEvoRec *pDevEvo = pDispEvo->pDevEvo;
     const NVDispHeadStateEvoRec *pHeadState =
                                  &pDispEvo->headState[head];
     NVT_HDR_INFOFRAME hdrInfoFrame = { 0 };
-    NVHDMIPKT_TC transmitControl;
     const NVT_HDR_STATIC_METADATA *pHdrInfo =
         &pEdidInfo->hdr_static_metadata_info;
+    NvEvoInfoFrameTransmitControl transmitCtrl =
+        NV_EVO_INFOFRAME_TRANSMIT_CONTROL_INIT;
 
     // Only send the HDMI HDR infoframe if the display supports HDR
     if (!pHdrInfo->supported_eotf.smpte_st_2084_eotf ||
@@ -648,7 +504,6 @@ SendHDRInfoFrame(const NVDispEvoRec *pDispEvo, const NvU32 head,
 
     if (pHeadState->hdrInfoFrame.state == NVKMS_HDR_INFOFRAME_STATE_ENABLED) {
         hdrInfoFrame.payload.eotf = pHeadState->hdrInfoFrame.eotf;
-
         hdrInfoFrame.payload.static_metadata_desc_id =
             NVT_CEA861_STATIC_METADATA_SM0;
 
@@ -659,25 +514,28 @@ SendHDRInfoFrame(const NVDispEvoRec *pDispEvo, const NvU32 head,
                      (const NvU16 *) &pHeadState->hdrInfoFrame.staticMetadata,
                      sizeof(NVT_HDR_INFOFRAME_MASTERING_DATA));
 
-        transmitControl = NVHDMIPKT_TRANSMIT_CONTROL_ENABLE_EVERY_FRAME;
+        transmitCtrl = NV_EVO_INFOFRAME_TRANSMIT_CONTROL_EVERY_FRAME;
     } else if (pHeadState->hdrInfoFrame.state ==
                NVKMS_HDR_INFOFRAME_STATE_TRANSITIONING) {
         nvDpyAssignSDRInfoFramePayload(&hdrInfoFrame.payload);
 
-        transmitControl = NVHDMIPKT_TRANSMIT_CONTROL_ENABLE_EVERY_FRAME;
+        transmitCtrl = NV_EVO_INFOFRAME_TRANSMIT_CONTROL_EVERY_FRAME;
     } else {
         nvAssert(pHeadState->hdrInfoFrame.state == NVKMS_HDR_INFOFRAME_STATE_DISABLED);
 
         nvDpyAssignSDRInfoFramePayload(&hdrInfoFrame.payload);
 
-        transmitControl = NVHDMIPKT_TRANSMIT_CONTROL_ENABLE_SINGLE_FRAME;
+        transmitCtrl = NV_EVO_INFOFRAME_TRANSMIT_CONTROL_SINGLE_FRAME;
     }
 
-    SendInfoFrame(pDispEvo,
-                  head,
-                  transmitControl,
-                  (NVT_INFOFRAME_HEADER *) &hdrInfoFrame.header,
-                  sizeof(hdrInfoFrame));
+    pDevEvo->hal->SendHdmiInfoFrame(
+        pDispEvo,
+        head,
+        transmitCtrl,
+        (NVT_INFOFRAME_HEADER *) &hdrInfoFrame.header,
+        (/* header length */ sizeof(hdrInfoFrame.header) +
+         /* payload length */ hdrInfoFrame.header.length),
+        TRUE /* needChecksum */);
 }
 
 
@@ -707,9 +565,10 @@ void nvUpdateHdmiInfoFrames(const NVDispEvoRec *pDispEvo,
                        pInfoFrameState,
                        &pDpyEvo->parsedEdid.info);
 
-    SendHDMI3DVendorSpecificInfoFrame(pDispEvo,
-                                      head,
-                                      &pDpyEvo->parsedEdid.info);
+    SendVendorSpecificInfoFrame(pDispEvo,
+                                head,
+                                pInfoFrameState,
+                                &pDpyEvo->parsedEdid.info);
 
     SendHDRInfoFrame(pDispEvo,
                      head,
@@ -1401,7 +1260,8 @@ void nvDpyUpdateHdmiVRRCaps(NVDpyEvoPtr pDpyEvo)
 
     const NvBool gpuSupportsHDMIVRR = pDevEvo->hal->caps.supportsHDMIVRR;
 
-    const NvBool dispSupportsVrr = nvDispSupportsVrr(pDispEvo);
+    const NvBool dispSupportsVrr = nvDispSupportsVrr(pDispEvo) && 
+                                       !nvkms_conceal_vrr_caps();
 
     const NvU32 edidVrrMin = pParsedEdid->info.hdmiForumInfo.vrr_min;
 
@@ -1415,8 +1275,6 @@ void nvDpyUpdateHdmiVRRCaps(NVDpyEvoPtr pDpyEvo)
             pDpyEvo->vrr.type =
                 NVKMS_DPY_VRR_TYPE_ADAPTIVE_SYNC_NON_DEFAULTLISTED;
         }
-
-        pDpyEvo->vrr.needsSwFramePacing = TRUE;
     }
 }
 
@@ -1841,9 +1699,12 @@ NvBool nvHdmi204k60HzRGB444Allowed(const NVDpyEvoRec *pDpyEvo,
  */
 void nvHdmiSetVRR(NVDispEvoPtr pDispEvo, NvU32 head, NvBool enable)
 {
+    const NVDevEvoRec *pDevEvo = pDispEvo->pDevEvo;
     NVT_EXTENDED_METADATA_PACKET_INFOFRAME empInfoFrame;
     NVT_EXTENDED_METADATA_PACKET_INFOFRAME_CTRL empCtrl = { 0 };
-    NVHDMIPKT_TC transmitControl;
+    NvEvoInfoFrameTransmitControl transmitCtrl = enable ?
+        NV_EVO_INFOFRAME_TRANSMIT_CONTROL_EVERY_FRAME :
+        NV_EVO_INFOFRAME_TRANSMIT_CONTROL_SINGLE_FRAME;
     NVT_STATUS status;
 
     empCtrl.EnableVRR = enable;
@@ -1857,24 +1718,14 @@ void nvHdmiSetVRR(NVDispEvoPtr pDispEvo, NvU32 head, NvBool enable)
         return;
     }
 
-    transmitControl =
-        DRF_DEF(_HDMI_PKT, _TRANSMIT_CTRL, _ENABLE,    _EN)   |
-        DRF_DEF(_HDMI_PKT, _TRANSMIT_CTRL, _OTHER,     _DIS)  |
-        DRF_DEF(_HDMI_PKT, _TRANSMIT_CTRL, _CHKSUM_HW, _DIS);
-
-    // Transmit the enable packet every frame, but only transmit the
-    // disable packet once.
-    if (enable) {
-        transmitControl |= DRF_DEF(_HDMI_PKT, _TRANSMIT_CTRL, _SINGLE, _DIS);
-    } else {
-        transmitControl |= DRF_DEF(_HDMI_PKT, _TRANSMIT_CTRL, _SINGLE, _EN);
-    }
-
-    SendInfoFrame(pDispEvo,
-                  head,
-                  transmitControl,
-                  (NVT_INFOFRAME_HEADER *) &empInfoFrame,
-                  sizeof(empInfoFrame));
+    // XXX Extended metadata infoframes do not contain a length header field.
+    pDevEvo->hal->SendHdmiInfoFrame(
+        pDispEvo,
+        head,
+        transmitCtrl,
+        (NVT_INFOFRAME_HEADER *) &empInfoFrame,
+        sizeof(empInfoFrame),
+        FALSE /* needChecksum */);
 }
 
 /*
@@ -2028,14 +1879,34 @@ static void HdmiLibPrint(
 }
 
 static void HdmiLibAssert(
-    NvHdmiPkt_CBHandle handle,
-    NvBool expr)
+    const char *expr,
+    const char *filename,
+    const char *function,
+    unsigned int line)
 {
-    /*
-     * This interface isn't the best... I hope you have a kernel debugger if
-     * this fires, because the file and line number will always be this one.
-     */
-    nvAssert(expr);
+#ifdef DEBUG
+    nvDebugAssert(expr, filename, function, line);
+#endif
+}
+
+static NvU64 hdmiLibTimerStartTime = 0;
+static NvU64 hdmiLibTimerTimeout = 0;
+
+static NvBool HdmiLibSetTimeout(NvHdmiPkt_CBHandle handle,
+                                NvU32 timeoutUs)
+{
+    hdmiLibTimerTimeout = timeoutUs;
+    hdmiLibTimerStartTime = nvkms_get_usec();
+    return TRUE;
+}
+
+static NvBool HdmiLibCheckTimeout(NvHdmiPkt_CBHandle handle)
+{
+    const NvU64 currentTime = nvkms_get_usec();
+    if (currentTime < hdmiLibTimerStartTime) {
+        return TRUE;
+    }
+    return (currentTime - hdmiLibTimerStartTime) > hdmiLibTimerTimeout;
 }
 
 static const NVHDMIPKT_CALLBACK HdmiLibCallbacks =
@@ -2045,8 +1916,8 @@ static const NVHDMIPKT_CALLBACK HdmiLibCallbacks =
     .rmDispControl2 = HdmiLibRmDispControl,
     .acquireMutex = HdmiLibAcquireMutex,
     .releaseMutex = HdmiLibReleaseMutex,
-    .setTimeout = NULL,     /* optional */
-    .checkTimeout = NULL,   /* optional */
+    .setTimeout = HdmiLibSetTimeout,
+    .checkTimeout = HdmiLibCheckTimeout,
     .malloc = HdmiLibMalloc,
     .free = HdmiLibFree,
     .print = HdmiLibPrint,
@@ -2119,7 +1990,6 @@ NvBool nvHdmiFrlAssessLink(NVDpyEvoPtr pDpyEvo)
  */
 NvBool nvHdmiDpySupportsFrl(const NVDpyEvoRec *pDpyEvo)
 {
-    NvU32 passiveDpDongleMaxPclkKHz;
     const NVDevEvoRec *pDevEvo = pDpyEvo->pDispEvo->pDevEvo;
 
     nvAssert(nvDpyIsHdmiEvo(pDpyEvo));
@@ -2134,45 +2004,87 @@ NvBool nvHdmiDpySupportsFrl(const NVDpyEvoRec *pDpyEvo)
         return FALSE;
     }
 
+    /*
+     * Can't use FRL if the connector is not natively HDMI (e.g., if
+     * using a passive DP-to-HDMI dongle, or if overrideEdid/forceConnected
+     * attempted to force HDMI FRL on a DP connector).
+     */
+    if (pDpyEvo->pConnectorEvo->type != NVKMS_CONNECTOR_TYPE_HDMI) {
+        return FALSE;
+    }
+
     /* Can't use FRL if the HDMI sink doesn't support it. */
     if (!pDpyEvo->parsedEdid.valid ||
         !pDpyEvo->parsedEdid.info.hdmiForumInfo.max_FRL_Rate) {
         return FALSE;
     }
 
-    /* Can't use FRL if we are using a passive DP to HDMI dongle. */
-    if (nvDpyGetPassiveDpDongleType(pDpyEvo, &passiveDpDongleMaxPclkKHz) !=
-        NV_EVO_PASSIVE_DP_DONGLE_UNUSED) {
-        return FALSE;
-    }
-
     return TRUE;
 }
 
-NvU32 nvHdmiGetEffectivePixelClockKHz(const NVDpyEvoRec *pDpyEvo,
-                                       const NVHwModeTimingsEvo *pHwTimings,
-                                       const NVDpyAttributeColor *pDpyColor)
+NvBool nvHdmiIsTmdsPossible(const NVDpyEvoRec *pDpyEvo,
+                            const NVHwModeTimingsEvo *pHwTimings,
+                            const NVDpyAttributeColor *pDpyColor)
 {
+    /* For YUV420 HW mode, divide pixel clock by 2. */
     const NvU32 pixelClock = (pHwTimings->yuv420Mode == NV_YUV420_MODE_HW) ?
         (pHwTimings->pixelClock / 2) : pHwTimings->pixelClock;
 
+    ct_assert(NV_KMS_DPY_ATTRIBUTE_CURRENT_COLOR_BPC_10 ==
+              NV_KMS_DPY_ATTRIBUTE_CURRENT_COLOR_BPC_MAX);
+
     nvAssert((pHwTimings->yuv420Mode == NV_YUV420_MODE_NONE) ||
                 (pDpyColor->format ==
                  NV_KMS_DPY_ATTRIBUTE_CURRENT_COLOR_SPACE_YCbCr420));
     nvAssert(nvDpyIsHdmiEvo(pDpyEvo));
+
+    /*
+     * HDMI requires 8+ BPC, enforced via nvDpyGetOutputColorFormatInfo() and
+     * IsColorBpcSupported(), so just assert instead of handling < 8 BPC.
+     */
     nvAssert(pDpyColor->bpc >= NV_KMS_DPY_ATTRIBUTE_CURRENT_COLOR_BPC_8);
 
-    /* YCbCr422 does not change the effective pixel clock. */
-    if (pDpyColor->format ==
-            NV_KMS_DPY_ATTRIBUTE_CURRENT_COLOR_SPACE_YCbCr422) {
-        return pixelClock;
+    /* For YCbCr422, compare without adjusting maximum pixel clock despite BPC. */
+    if (pDpyColor->format == NV_KMS_DPY_ATTRIBUTE_CURRENT_COLOR_SPACE_YCbCr422) {
+        return (pixelClock <= pDpyEvo->maxSingleLinkPixelClockKHz);
     }
 
     /*
-     * For > 8 BPC, the effective pixel clock is adjusted upwards according to
-     * the ratio of the given BPC and 8 BPC.
+     * For 10 BPC, compare with hardware reduced limit if applicable, otherwise
+     * adjust maximum pixel clock by a ratio of 8/10 BPC.
      */
-    return ((pixelClock * pDpyColor->bpc) / 8ULL);
+    if (pDpyColor->bpc == NV_KMS_DPY_ATTRIBUTE_CURRENT_COLOR_BPC_10) {
+        NvU32 hdmiTmds10BpcMaxPClkKHz =
+            pDpyEvo->pDispEvo->pDevEvo->caps.hdmiTmds10BpcMaxPClkMHz * 1000UL;
+        NvU32 adjustedMaxPixelClock =
+            (pDpyEvo->maxSingleLinkPixelClockKHz * 4ULL) / 5ULL;
+        NvU32 adjustedMaxEDIDPixelClock =
+            pDpyEvo->parsedEdid.valid ?
+              (pDpyEvo->parsedEdid.limits.max_pclk_10khz * 10 * 4ULL) / 5ULL : 0;
+
+        /* Pixel clock must satisfy hdmiTmds10BpcMaxPClkKHz, if applicable. */
+        if ((hdmiTmds10BpcMaxPClkKHz > 0) &&
+            (pixelClock > hdmiTmds10BpcMaxPClkKHz)) {
+            return FALSE;
+        }
+
+        /* Pixel clock must also satisfy adjustedMaxPixelClock. */
+        if (pixelClock > adjustedMaxPixelClock) {
+            return FALSE;
+        }
+
+        /* Pixel clock must also satisfy adjustedMaxEDIDPixelClock. */
+        if (adjustedMaxEDIDPixelClock != 0 &&
+            pixelClock > adjustedMaxEDIDPixelClock) {
+            return FALSE;
+        }
+
+        return TRUE;
+    }
+
+    /* For 8 BPC, compare without adjusting maximum pixel clock. */
+    nvAssert(pDpyColor->bpc == NV_KMS_DPY_ATTRIBUTE_CURRENT_COLOR_BPC_8);
+    return (pixelClock <= pDpyEvo->maxSingleLinkPixelClockKHz);
 }
 
 static NvU64 GetHdmiFrlLinkRate(HDMI_FRL_DATA_RATE frlRate)
@@ -2234,8 +2146,9 @@ static NvBool nvHdmiFrlQueryConfigOneBpc(
 
     nvAssert(nvDpyIsHdmiEvo(pDpyEvo));
     nvAssert(nvHdmiDpySupportsFrl(pDpyEvo));
-    nvAssert(nvHdmiGetEffectivePixelClockKHz(pDpyEvo, pHwTimings, pDpyColor) >
-                pDpyEvo->maxSingleLinkPixelClockKHz);
+
+    nvAssert(!nvHdmiIsTmdsPossible(pDpyEvo, pHwTimings, pDpyColor) ||
+             nvGetPreferHdmiFrlMode(pDevEvo, pValidationParams));
 
     /* See if we can find an NVT_TIMING for this mode from the EDID. */
     pNvtTiming = nvFindEdidNVT_TIMING(pDpyEvo, pModeTimings, pValidationParams);
@@ -2368,6 +2281,16 @@ static NvBool nvHdmiFrlQueryConfigOneBpc(
                 }
 
                 pDscInfo->type = NV_DSC_INFO_EVO_TYPE_HDMI;
+                pDscInfo->sliceCount = pConfig->dscInfo.sliceCount;
+                /*
+                 * XXX NvHdmi_QueryFRLConfig() might get updated in future, but
+                 * today it does not return the possible DSC slice counts.
+                 * NvHdmi_QueryFRLConfig() returns the slice count which it has
+                 * been used to calculate DSC PPS, populate
+                 * 'possibleSliceCountMask' using that slice count.
+                 */
+                pDscInfo->possibleSliceCountMask =
+                    NVBIT(pDscInfo->sliceCount - 1);
                 pDscInfo->hdmi.dscMode = b2Heads1Or ?
                     NV_DSC_EVO_MODE_DUAL : NV_DSC_EVO_MODE_SINGLE;
                 pDscInfo->hdmi.bitsPerPixelX16 =
